@@ -5,8 +5,11 @@
 #include <list>
 #include <optional>
 #include <iterator>
+#include <fstream>
+#include <string>
 
 // MATA headers
+#include "mata/alphabet.hh"
 #include "mata/utils/sparse-set.hh"
 #include "mata/nfa/nfa.hh"
 #include "mata/nfa/algorithms.hh"
@@ -117,6 +120,25 @@ std::vector<State> Nfa::distances_from_initial() const {
     return distances;
 }
 
+std::vector<State> Nfa::distances_to_final() const {
+    return revert(*this).distances_from_initial();
+}
+
+Run Nfa::get_shortest_accepting_run_from_state(State q, const std::vector<State>& distances_to_final) const {
+    Run result{{}, {q}};
+    while (!final[q]) {
+        for (Move move : delta[q].moves()) {
+            if (distances_to_final[move.target] < distances_to_final[q]) {
+                result.word.push_back(move.symbol);
+                result.path.push_back(move.target);
+                q = move.target;
+                break;
+            }
+        }
+    }
+    return result;
+}
+
 Nfa& Nfa::trim(StateRenaming* state_renaming) {
 #ifdef _STATIC_STRUCTURES_
     BoolVector useful_states{ useful_states() };
@@ -385,7 +407,7 @@ bool Nfa::is_flat() const {
     mata::nfa::Nfa::TarjanDiscoverCallback callback {};
     callback.scc_discover = [&](const std::vector<mata::nfa::State>& scc, const std::vector<mata::nfa::State>& tarjan_stack) -> bool {
         (void)tarjan_stack;
-        
+
         for(const mata::nfa::State& st : scc) {
             bool one_input_visited = false;
             for (const mata::nfa::SymbolPost& sp : this->delta[st]) {
@@ -407,31 +429,128 @@ bool Nfa::is_flat() const {
     return flat;
 }
 
-std::string Nfa::print_to_dot() const {
+std::string Nfa::print_to_dot(const bool decode_ascii_chars, const bool use_intervals, const int max_label_length) const {
     std::stringstream output;
-    print_to_dot(output);
+    print_to_dot(output, decode_ascii_chars, use_intervals, max_label_length);
     return output.str();
 }
 
-void Nfa::print_to_dot(std::ostream &output) const {
+void Nfa::print_to_dot(std::ostream &output, const bool decode_ascii_chars, const bool use_intervals, const int max_label_length) const {
+    auto to_ascii = [&](const Symbol symbol) -> std::string {
+        // Translate only printable ASCII characters.
+        if (symbol < 33 || symbol >= 127) {
+            return "<" + std::to_string(symbol) + ">";
+        }
+        switch (symbol) {
+            case '"':     return "\\\"";
+            case '\\':    return "\\\\";
+            default:      return std::string(1, static_cast<char>(symbol));
+        }
+    };
+
+    auto translate_symbol = [&](const Symbol symbol) -> std::string {
+        switch (symbol) {
+            case EPSILON:      return "<eps>";
+            default:           break;
+        }
+        if (decode_ascii_chars) {
+            return to_ascii(symbol);
+        }
+        return std::to_string(symbol);
+    };
+
+    auto vec_of_symbols_to_string = [&](const OrdVector<Symbol>& symbols) {
+        std::string result;
+        for (const Symbol& symbol: symbols) {
+            result += translate_symbol(symbol) + ",";
+        }
+        result.pop_back(); // Remove last comma
+        return result;
+    };
+
+    auto vec_of_symbols_to_string_with_intervals = [&](const OrdVector<Symbol>& symbols) {
+        std::string result;
+
+        std::vector<std::pair<Symbol, Symbol>> intervals;
+        auto symbols_it = symbols.begin();
+        std::pair<Symbol, Symbol> interval{*symbols_it, *symbols_it};
+        ++symbols_it;
+        for (; symbols_it != symbols.end(); ++symbols_it) {
+            if (*symbols_it == interval.second + 1) {
+                interval.second = *symbols_it;
+            } else {
+                intervals.push_back(interval);
+                interval = {*symbols_it, *symbols_it};
+            }
+        }
+        intervals.push_back(interval);
+
+        for (const auto& interval: intervals) {
+            const size_t interval_size = interval.second - interval.first + 1;
+            if (interval_size == 1) {
+                result += translate_symbol(interval.first) + ",";
+            } else if (interval_size == 2) {
+                result += translate_symbol(interval.first) + "," + translate_symbol(interval.second) + ",";
+            } else {
+                result += "[" + translate_symbol(interval.first) + "-" + translate_symbol(interval.second) + "],";
+            }
+        }
+
+        result.pop_back(); // Remove last comma
+        return result;
+    };
+
+    BoolVector is_state_drawn(num_of_states(), false);
     output << "digraph finiteAutomaton {" << std::endl
                  << "node [shape=circle];" << std::endl;
 
+    // Double circle for final states
     for (State final_state: final) {
+        is_state_drawn[final_state] = true;
         output << final_state << " [shape=doublecircle];" << std::endl;
     }
 
+    // Print transitions
     const size_t delta_size = delta.num_of_states();
     for (State source = 0; source != delta_size; ++source) {
+        std::unordered_map<State, OrdVector<Symbol>> tgt_symbols_map;
         for (const SymbolPost &move: delta[source]) {
-            output << source << " -> {";
+            is_state_drawn[source] = true;
             for (State target: move.targets) {
-                output << target << " ";
+                is_state_drawn[target] = true;
+                tgt_symbols_map[target].insert(move.symbol);
+            }
+        }
+        for (const auto& [target, symbols]: tgt_symbols_map) {
+            if (max_label_length == 0) {
+                output << source << " -> " << target << ";" << std::endl;
+                continue;
+            }
+
+            std::string label = (use_intervals) ? vec_of_symbols_to_string_with_intervals(symbols) : vec_of_symbols_to_string(symbols);
+            std::string on_hover_label = utils::replace_all(utils::replace_all(label, "<", "&lt;"), ">", "&gt;");
+            bool is_shortened = false;
+            if (max_label_length > 0 && label.length() > static_cast<size_t>(max_label_length)) {
+                label.replace(static_cast<size_t>(max_label_length), std::string::npos, "...");
+                is_shortened = true;
+            }
+
+            if (is_shortened) {
+                output << source << " -> " << target << " [label=\"" << label << "\", tooltip=\"" << on_hover_label << "\"];" << std::endl;
+            } else {
+                output << source << " -> " << target << " [label=\"" << label << "\"];" << std::endl;
             }
-            output << "} [label=" << move.symbol << "];" << std::endl;
         }
     }
 
+    // Circle for isolated states with no transitions
+    for (State state{ 0 }; state < is_state_drawn.size(); ++state) {
+        if (!is_state_drawn[state]) {
+            output << state << " [shape=circle];" << std::endl;
+        }
+    }
+
+    // Arrow for initial states
     output << "node [shape=none, label=\"\"];" << std::endl;
     for (State init_state: initial) {
         output << "i" << init_state << " -> " << init_state << ";" << std::endl;
@@ -440,13 +559,21 @@ void Nfa::print_to_dot(std::ostream &output) const {
     output << "}" << std::endl;
 }
 
-std::string Nfa::print_to_mata() const {
+void Nfa::print_to_dot(const std::string& filename, const bool decode_ascii_chars, const bool use_intervals, const int max_label_length) const {
+    std::ofstream output(filename);
+    if (!output) {
+        throw std::ios_base::failure("Failed to open file: " + filename);
+    }
+    print_to_dot(output, decode_ascii_chars, use_intervals, max_label_length);
+}
+
+std::string Nfa::print_to_mata(const Alphabet* alphabet) const {
     std::stringstream output;
-    print_to_mata(output);
+    print_to_mata(output, alphabet);
     return output.str();
 }
 
-void Nfa::print_to_mata(std::ostream &output) const {
+void Nfa::print_to_mata(std::ostream &output, const Alphabet* alphabet) const {
     output << "@NFA-explicit" << std::endl
            << "%Alphabet-auto" << std::endl;
            // TODO should be this, but we cannot parse %Alphabet-numbers yet
@@ -469,12 +596,22 @@ void Nfa::print_to_mata(std::ostream &output) const {
     }
 
     for (const Transition& trans: delta.transitions()) {
-        output << "q" << trans.source << " " << trans.symbol << " q" << trans.target << std::endl;
+        output << "q" << trans.source << " "
+        << ((alphabet != nullptr) ? alphabet->reverse_translate_symbol(trans.symbol) : std::to_string(trans.symbol))
+        << " q" << trans.target << std::endl;
     }
 }
 
+void Nfa::print_to_mata(const std::string& filename, const Alphabet* alphabet) const {
+    std::ofstream output(filename);
+    if (!output) {
+        throw std::ios_base::failure("Failed to open file: " + filename);
+    }
+    print_to_mata(output, alphabet);
+}
+
 Nfa Nfa::get_one_letter_aut(Symbol abstract_symbol) const {
-    Nfa digraph{num_of_states(), StateSet(initial), StateSet(final) };
+    Nfa digraph{num_of_states(), initial, final };
     // Add directed transitions for digraph.
     for (const Transition& transition: delta.transitions()) {
         // Directly try to add the transition. Finding out whether the transition is already in the digraph
@@ -488,51 +625,97 @@ void Nfa::get_one_letter_aut(Nfa& result) const {
     result = get_one_letter_aut();
 }
 
-StateSet Nfa::post(const StateSet& states, const Symbol& symbol) const {
+StateSet Nfa::post(const StateSet& states, const Symbol symbol, const EpsilonClosureOpt epsilon_closure_opt) const {
+    auto get_epsilon_closure = [&](const StateSet& states) {
+        StateSet closure{ states };
+        std::queue<State> worklist;
+        for (const State state: states) {
+            worklist.push(state);
+        }
+        while (!worklist.empty()) {
+            const State state = worklist.front();
+            worklist.pop();
+            auto move_it{ delta[state].find(EPSILON) };
+            if (move_it != delta[state].end()) {
+                for (const State target: move_it->targets) {
+                    if (!closure.contains(target)) {
+                        closure.insert(target);
+                        worklist.push(target);
+                    }
+                }
+            }
+        }
+        return closure;
+    };
+
     StateSet res{};
+
+    // If the symbol is EPSILON, we can stay in the same state.
+    if (symbol == EPSILON && epsilon_closure_opt != EpsilonClosureOpt::NONE) {
+        res = states;
+    }
+
     if (delta.empty()) {
         return res;
     }
 
-    for (const State state: states) {
+    StateSet from_states = states;
+    if (epsilon_closure_opt == EpsilonClosureOpt::BEFORE) {
+        // Before making the step using the symbol, we compute the epsilon closure.
+        from_states = get_epsilon_closure(states);
+    }
+
+    // Now, we can make the step using the symbol.
+    for (const State state: from_states) {
         const StatePost& post{ delta[state] };
-        // TODO: This does not handle epsilons.
         const auto move_it{ post.find(symbol) };
         if (move_it != post.end()) {
             res.insert(move_it->targets);
         }
     }
+
+    if (epsilon_closure_opt == EpsilonClosureOpt::AFTER) {
+        // We need to compute the epsilon closure of the resulting states.
+        res = get_epsilon_closure(res);
+    }
+
     return res;
 }
 
- void Nfa::unify_initial() {
-    if (initial.empty() || initial.size() == 1) { return; }
-    const State new_initial_state{add_state() };
+ Nfa& Nfa::unify_initial(const bool force_new_state) {
+    if (!force_new_state && (initial.empty() || initial.size() == 1)) { return *this; }
+
+    const State new_initial_state{ add_state() };
     for (const State orig_initial_state: initial) {
-        const StatePost& moves{ delta.state_post(orig_initial_state) };
-        for (const auto& transitions: moves) {
-            for (const State state_to: transitions.targets) {
-                delta.add(new_initial_state, transitions.symbol, state_to);
+        const StatePost& state_post{ delta.state_post(orig_initial_state) };
+        for (const auto& symbol_post: state_post) {
+            for (const State target: symbol_post.targets) {
+                delta.add(new_initial_state, symbol_post.symbol, target);
             }
         }
         if (final[orig_initial_state]) { final.insert(new_initial_state); }
     }
+
     initial.clear();
     initial.insert(new_initial_state);
+    return *this;
 }
 
-void Nfa::unify_final() {
-    if (final.empty() || final.size() == 1) { return; }
+Nfa& Nfa::unify_final(const bool force_new_state) {
+    if (!force_new_state && (final.empty() || final.size() == 1)) { return *this; }
+
     const State new_final_state{ add_state() };
     for (const auto& orig_final_state: final) {
         const auto transitions_to{ delta.get_transitions_to(orig_final_state) };
-        for (const auto& transitions: transitions_to) {
-            delta.add(transitions.source, transitions.symbol, new_final_state);
+        for (const auto& transition: transitions_to) {
+            delta.add(transition.source, transition.symbol, new_final_state);
         }
         if (initial[orig_final_state]) { initial.insert(new_final_state); }
     }
+
     final.clear();
     final.insert(new_final_state);
+    return *this;
 }
 
 Nfa& Nfa::operator=(Nfa&& other) noexcept {
@@ -560,6 +743,36 @@ State Nfa::add_state(State state) {
     return state;
 }
 
+State Nfa::insert_word(const State source, const Word &word, const State target) {
+    assert(!word.empty());
+    assert(source < num_of_states());
+    assert(target < num_of_states());
+
+    const size_t word_len = word.size();
+    if (word_len == 1) {
+        delta.add(source, word[0], target);
+        return target;
+    }
+
+    // Add transition source --> inner_state.
+    State inner_state = add_state();
+    delta.add(source, word[0], inner_state);
+
+    // Add transitions inner_state --> inner_state
+    State prev_state = inner_state;
+    for (size_t idx{ 1 }; idx < word_len - 1; idx++) {
+        inner_state = add_state();
+        delta.add(prev_state, word[idx], inner_state);
+        prev_state = inner_state;
+    }
+
+    // Add transition inner_state --> target
+    delta.add(prev_state, word[word_len - 1], target);
+    return target;
+}
+
+State Nfa::insert_word(const State source, const Word &word) { return insert_word(source, word, add_state()); }
+
 size_t Nfa::num_of_states() const {
     return std::max({
         static_cast<size_t>(initial.domain_size()),
@@ -629,3 +842,112 @@ Nfa& Nfa::unite_nondet_with(const mata::nfa::Nfa& aut) {
 
     return *this;
 }
+
+Nfa Nfa::decode_utf8() const {
+    Nfa result{ num_of_states(), { initial }, { final } };
+    BoolVector used(num_of_states(), false);
+    std::stack<State> worklist;
+
+    // Pushes a set of states to the worklist and marks them as used.
+    auto push_state_set = [&](const StateSet& set) {
+        for (State state: set) {
+            if (used[state]) {
+                continue;
+            }
+            worklist.push(state);
+            used[state] = true;
+        }
+    };
+
+    // Adds a symbol_post to the state_post.
+    // If the transition sequence is deterministic, we can use emplace_back
+    // because symbols are discovered in ascending order. However, in cases
+    // of nondeterministic sequences, we must use insert to ensure proper ordering.
+    // For example, consider the sequences 0xC8 0x80 and 0xC8 0x88.
+    // Based solely on the first byte (0xC8), we cannot determine which sequence
+    // will result in the higher number.
+    auto add_to_state_post = [&](StatePost &state_post, const SymbolPost& symbol_post, const bool is_nondet) {
+        if (is_nondet) {
+            state_post.insert(symbol_post);
+        } else {
+            state_post.emplace_back(std::move(symbol_post));
+        }
+    };
+
+    // UTF-8 Byte Patterns:
+    // U+0000   to U+007F  : 0xxxxxxx
+    // U+0080   to U+07FF  : 110xxxxx 10xxxxxx
+    // U+0800   to U+FFFF  : 1110xxxx 10xxxxxx 10xxxxxx
+    // U+010000 to U+10FFFF: 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
+    // NOTE: Due to the nature of RE2, the automaton language can contain unexpected (invalid)
+    //       UTF-8 sequences, such as 11000000 10000000 (U+0300). Because of that,
+    //       we need to check if the decoded symbol is within the valid range of Unicode code points.
+    push_state_set(StateSet{this->initial});
+    while (!worklist.empty()) {
+        State q1 = worklist.top();
+        StatePost &q1_state_post = result.delta.mutable_state_post(q1);
+        worklist.pop();
+        // 1st Byte
+        for (const SymbolPost &sp1: this->delta[q1]) {
+            const Symbol s1 = sp1.symbol;
+            if ((s1 & 0x80) == 0x00) {
+                q1_state_post.emplace_back(SymbolPost{s1, sp1.targets});;
+                push_state_set(sp1.targets);
+                continue;
+            }
+            // 2nd Byte
+            const bool is_nondet1 = sp1.targets.size() > 1;
+            for (const State q2: sp1.targets) {
+                for (const SymbolPost &sp2: this->delta[q2]) {
+                    const Symbol s2 = sp2.symbol;
+                    if ((s1 & 0xE0) == 0xC0) {
+                        assert((s2 & 0xC0) == 0x80);
+                        const Symbol symbol = ((s1 & 0x1F) << 6) | (s2 & 0x3F);
+                        if (symbol < 0x80) {
+                            continue;   // Invalid UTF-8 sequence
+                        }
+                        assert(symbol <= 0x7FF);
+                        add_to_state_post(q1_state_post, SymbolPost{symbol, sp2.targets}, is_nondet1);
+                        push_state_set(sp2.targets);
+                        continue;
+                    }
+                    // 3rd Byte
+                    const bool is_nondet2 = is_nondet1 || sp2.targets.size() > 1;
+                    for (const State q3: sp2.targets) {
+                        for (const SymbolPost &sp3: this->delta[q3]) {
+                            const Symbol s3 = sp3.symbol;
+                            if ((s1 & 0xF0) == 0xE0) {
+                                assert((s3 & 0xC0) == 0x80);
+                                const Symbol symbol = ((s1 & 0x0F) << 12) | ((s2 & 0x3F) << 6) | (s3 & 0x3F);
+                                if (symbol < 0x800) {
+                                    continue;   // Invalid UTF-8 sequence
+                                }
+                                assert(symbol <= 0xFFFF);
+                                add_to_state_post(q1_state_post, SymbolPost{symbol, sp3.targets}, is_nondet2);
+                                push_state_set(sp3.targets);
+                                continue;
+                            }
+                            // 4th Byte
+                            const bool is_nondet3 = is_nondet2 || sp3.targets.size() > 1;
+                            for (const State q4: sp3.targets) {
+                                for (const SymbolPost &sp4: this->delta[q4]) {
+                                    const Symbol s4 = sp4.symbol;
+                                    assert((s1 & 0xF8) == 0xF0);
+                                    assert((s4 & 0xC0) == 0x80);
+                                    const Symbol symbol = ((s1 & 0x07) << 18) | ((s2 & 0x3F) << 12) | ((s3 & 0x3F) << 6) | (s4 & 0x3F);
+                                    if (symbol < 0x10000 || symbol > 0x10FFFF) {
+                                        continue;   // Invalid UTF-8 sequence
+                                    }
+                                    add_to_state_post(q1_state_post, SymbolPost{symbol, sp4.targets}, is_nondet3);
+                                    push_state_set(sp4.targets);
+                                }
+                            }
+                        }
+                    }
+                }
+            }
+        }
+    }
+
+    return result;
+}
