@@ -46,14 +46,15 @@ namespace {
         /**
          * Creates parsed regex (ie. Regexp*) from string regex_string
          * @param regex_string Regex to be parsed as a string
+         * @param encoding Encoding of the regex, default is Latin1
          * @return Parsed regex as RE2 Regexp*
          */
-        re2::Regexp* parse_regex_string(const std::string& regex_string) const {
+        re2::Regexp* parse_regex_string(const std::string& regex_string, const Encoding encoding = Encoding::Latin1) const {
             re2::RegexpStatus status;
 
             auto parsed_regex = re2::Regexp::Parse(
                     regex_string,
-                    static_cast<re2::Regexp::ParseFlags>(options.ParseFlags()),
+                    static_cast<re2::Regexp::ParseFlags>(options.ParseFlags() | static_cast<int>(encoding)),
                     &status);
             if (parsed_regex == nullptr) {
                 if (options.log_errors()) {
@@ -77,7 +78,6 @@ namespace {
             const auto prog_size = static_cast<size_t>(prog->size());
             // The same symbol in lowercase and uppercase is 32 symbols from each other in ASCII
             const int ascii_shift_value = 32;
-            int empty_flag;
             std::vector<mata::Symbol> symbols;
             Nfa explicit_nfa(prog_size);
 
@@ -136,11 +136,11 @@ namespace {
             this->outgoingEdges = std::vector<std::vector<std::pair<mata::Symbol, mata::nfa::State>>> (prog_size);
 
             // We traverse all the states and create corresponding states and edges in Nfa
-            for (mata::nfa::State current_state = start_state; current_state < prog_size; current_state++) {
-                re2::Prog::Inst *inst = prog->inst(static_cast<int>(current_state));
+            for (State re2_state = start_state; re2_state < prog_size; ++re2_state) {
+                re2::Prog::Inst *inst = prog->inst(static_cast<int>(re2_state));
                 // Every type of state can be final (due to epsilon transition), so we check it regardless of its type
-                if (this->state_cache.is_final_state[current_state]) {
-                    this->make_state_final(current_state, explicit_nfa);
+                 if (this->state_cache.is_final_state[re2_state] && inst->opcode() == re2::kInstMatch) {
+                    this->make_state_final(re2_state, explicit_nfa);
                 }
                 switch (inst->opcode()) {
                     default:
@@ -156,41 +156,13 @@ namespace {
                     case re2::kInstCapture:
                         if (use_epsilon) {
                             symbols.push_back(epsilon_value);
-                            this->create_explicit_nfa_transitions(current_state, inst, symbols, explicit_nfa, use_epsilon, epsilon_value);
+                            this->create_explicit_nfa_transitions(re2_state, inst, symbols, explicit_nfa, use_epsilon, epsilon_value);
                             symbols.clear();
                         }
                         break;
                     case re2::kInstEmptyWidth:
-                        empty_flag = static_cast<int>(inst->empty());
-                        // ^ - beginning of line
-                        if (empty_flag & re2::kEmptyBeginLine) {
-                            // TODO Symbol?
-                            symbols.push_back(300);
-                        }
-                        // $ - end of line
-                        if (empty_flag & re2::kEmptyEndLine) {
-                            // TODO Symbol?
-                            symbols.push_back(10);
-                        }
-                        // \A - beginning of text
-                        if (empty_flag & re2::kEmptyBeginText) {
-                            // TODO Symbol?
-                            symbols.push_back(301);
-                        }
-                        // \z - end of text
-                        if (empty_flag & re2::kEmptyEndText) {
-                            // TODO Symbol?
-                            symbols.push_back(302);
-                        }
-                        // \b - word boundary
-                        if (empty_flag & re2::kEmptyWordBoundary) {
-                            // TODO Symbol?
-                            symbols.push_back(303);
-                        }
-                        // \B - not \b
-                        if (empty_flag & re2::kEmptyNonWordBoundary) {
-                            // TODO Symbol?
-                            symbols.push_back(304);
+                        if (use_epsilon) {
+                            this->create_explicit_nfa_transitions(re2_state, inst, {epsilon_value}, explicit_nfa, use_epsilon, epsilon_value);
                         }
                         break;
                     // kInstByteRange represents states with a "byte range" on the outgoing transition(s)
@@ -202,19 +174,19 @@ namespace {
                                 symbols.push_back(symbol);
                                 // Foldcase causes RE2 to do a case-insensitive match, so transitions will be made for
                                 // both uppercase and lowercase symbols
-                                if (inst->foldcase()) {
+                                if (inst->foldcase() && symbol >= 'a' && symbol <= 'z') {
                                     symbols.push_back(symbol-ascii_shift_value);
                                 }
                             }
                         }
-                        this->create_explicit_nfa_transitions(current_state, inst, symbols, explicit_nfa, use_epsilon, epsilon_value);
+                        this->create_explicit_nfa_transitions(re2_state, inst, symbols, explicit_nfa, use_epsilon, epsilon_value);
 
                         if (!use_epsilon) {
                             // There is an epsilon transition to the currentState+1 we will need to copy transitions of
                             // the currentState+1 to the currentState.
-                            if (!this->state_cache.is_last[current_state]) {
-                                for (auto state: this->state_cache.state_mapping[current_state + 1]) {
-                                    copyEdgesFromTo.emplace_back(state, current_state);
+                            if (!this->state_cache.is_last[re2_state]) {
+                                for (auto state: this->state_cache.state_mapping[re2_state + 1]) {
+                                    copyEdgesFromTo.emplace_back(state, re2_state);
                                 }
                             }
                         }
@@ -252,7 +224,7 @@ namespace {
                     }
                 }
             }
-            RegexParser::renumber_states(output_nfa, prog_size, explicit_nfa);
+            *output_nfa = Nfa(explicit_nfa).trim();
         }
 
     private: // private methods
@@ -419,7 +391,8 @@ namespace {
                 if (inst->last()) {
                     this->state_cache.is_last[state] = true;
                 }
-                if (inst->opcode() == re2::kInstMatch) {
+                if (inst->opcode() == re2::kInstMatch ||
+                    (inst->opcode() == re2::kInstEmptyWidth && inst->empty() & re2::kEmptyEndText)) {
                     this->state_cache.is_final_state[state] = true;
                 }
             }
@@ -440,58 +413,6 @@ namespace {
             }
         }
 
-        /**
-         * Renumbers the states of the input_nfa to be from <0, numberOfStates>
-         * @param program_size Size of the RE2 prog
-         * @param input_nfa Nfa which states should be renumbered
-         * @return Same Nfa as input_nfa but with states from interval <0, numberOfStates>
-         */
-        static Nfa renumber_states(Nfa* output_nfa,
-                                              size_t program_size,
-                                              Nfa &input_nfa) {
-            std::vector<mata::nfa::State> renumbered_states(program_size, mata::nfa::Limits::max_state);
-            Nfa& renumbered_explicit_nfa = *output_nfa;
-            for (mata::nfa::State state{ 0 }; state < program_size; state++) {
-                const auto& transition_list = input_nfa.delta.state_post(state);
-                // If the transition list is empty, the state is not used
-                if (transition_list.empty()) {
-                    continue;
-                } else {
-                    // addNewState returns next unused state of the new NFA, so we map it to the original state
-                    renumbered_states[state] = renumbered_explicit_nfa.add_state();
-                }
-            }
-
-            for (auto state: input_nfa.final) {
-                if (static_cast<int>(renumbered_states[state]) == -1) {
-                    renumbered_states[state] = renumbered_explicit_nfa.add_state();
-                }
-                renumbered_explicit_nfa.final.insert(renumbered_states[state]);
-            }
-
-            for (mata::nfa::State state{ 0 }; state < program_size; state++) {
-                const auto& transition_list = input_nfa.delta.state_post(state);
-                for (const auto& transition: transition_list) {
-                    for (auto stateTo: transition.targets) {
-                        if (renumbered_states[stateTo] == mata::nfa::Limits::max_state) {
-                            renumbered_states[stateTo] = renumbered_explicit_nfa.add_state();
-                        }
-                        assert(renumbered_states[state] <= renumbered_explicit_nfa.num_of_states());
-                        assert(renumbered_states[stateTo] <= renumbered_explicit_nfa.num_of_states());
-                        renumbered_explicit_nfa.delta.add(renumbered_states[state], transition.symbol,
-                                                          renumbered_states[stateTo]);
-                    }
-                }
-            }
-
-
-            for (auto state: input_nfa.initial) {
-                renumbered_explicit_nfa.initial.insert(renumbered_states[state]);
-            }
-
-            return renumbered_explicit_nfa;
-        }
-
         /**
          * Gets all states that are mapped to the state (i.e., states that are within epsilon transitions chain)
          * @param prog RE2 prog corresponding to the parsed regex
@@ -541,33 +462,31 @@ namespace {
         };
 }
 
- /**
- * The main method, it creates NFA from regex
- * @param pattern regex as string
- * @param use_epsilon whether to create NFA with epsilon transitions or not
- * @param epsilon_value value, that will represent epsilon on transitions
- * @param use_reduce if set to true the result is trimmed and reduced using simulation reduction
- * @return Nfa corresponding to pattern
- */
-void mata::parser::create_nfa(nfa::Nfa* nfa, const std::string& pattern, bool use_epsilon, mata::Symbol epsilon_value, bool use_reduce) {
-    if (nfa == nullptr) {
-        throw std::runtime_error("create_nfa: nfa should not be NULL");
-    }
-
+mata::nfa::Nfa mata::parser::create_nfa(const std::string& pattern, bool use_epsilon, mata::Symbol epsilon_value, bool use_reduce, const Encoding encoding) {
+    mata::nfa::Nfa result;
     RegexParser regexParser{};
-    auto parsed_regex = regexParser.parse_regex_string(pattern);
+    auto parsed_regex = regexParser.parse_regex_string(pattern, encoding);
     auto program = parsed_regex->CompileToProg(regexParser.options.max_mem() * 2 / 3);
-    regexParser.convert_pro_to_nfa(nfa, program, true, epsilon_value);
+    // FIXME: use_epsilon = false completely breaks the method convert_pro_to_nfa(). Needs fixing before allowing to
+    //  pass the argument use_epsilon to convert_pro_to_nfa().
+    regexParser.convert_pro_to_nfa(&result, program, true, epsilon_value);
     delete program;
     // Decrements reference count and deletes object if the count reaches 0
     parsed_regex->Decref();
+
      //TODO: should this really be done implicitly?
     if(!use_epsilon) {
-        *nfa = mata::nfa::remove_epsilon(*nfa, epsilon_value);
+        result = mata::nfa::remove_epsilon(result, epsilon_value);
     }
     //TODO: in fact, maybe parser should not do trimming and reducing, maybe these operations should be done transparently.
     if(use_reduce) {
         //TODO: trimming might be unnecessary, regex->nfa construction should not produce useless states. Or does it?
-        *nfa = mata::nfa::reduce(nfa->trim());
+        result = mata::nfa::reduce(result.trim());
     }
+
+    return result;
+}
+
+void mata::parser::create_nfa(nfa::Nfa* nfa, const std::string& pattern, bool use_epsilon, mata::Symbol epsilon_value, bool use_reduce, const Encoding encoding) {
+    *nfa = create_nfa(pattern, use_epsilon, epsilon_value, use_reduce, encoding);
 }
